@@ -1,5 +1,5 @@
 import itertools
-from typing import Dict, List
+from typing import Dict, List, Tuple
 
 import numpy as np
 from netsquid.components.instructions import (
@@ -17,11 +17,14 @@ from netsquid.components.instructions import (
     INSTR_Y,
     INSTR_Z,
 )
-from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel
+from netsquid.components.models.qerrormodels import (
+    DepolarNoiseModel,
+    QuantumErrorModel,
+    T1T2NoiseModel,
+)
 from netsquid.components.qprocessor import PhysicalInstruction, QuantumProcessor
 from netsquid.qubits.operators import Operator
 from netsquid_magic.link_layer import (
-    MagicLinkLayerProtocol,
     MagicLinkLayerProtocolWithSignaling,
     SingleClickTranslationUnit,
 )
@@ -30,27 +33,95 @@ from netsquid_magic.magic_distributor import (
     DoubleClickMagicDistributor,
     PerfectStateMagicDistributor,
 )
-from netsquid_nv.magic_distributor import NVSingleClickMagicDistributor
 from netsquid_physlayer.heralded_connection import MiddleHeraldedConnection
 
-from qoala.runtime.config import (
-    DepolariseLinkConfig,
+from qoala.lang.ehi import EhiLinkInfo, EhiNetworkInfo
+
+# Ignore type since whole 'config' module is ignored by mypy
+from qoala.runtime.config import (  # type: ignore
+    DepolariseOldLinkConfig,
     GenericQDeviceConfig,
-    HeraldedLinkConfig,
+    HeraldedOldLinkConfig,
     LinkConfig,
-    NVLinkConfig,
     NVQDeviceConfig,
     ProcNodeConfig,
     ProcNodeNetworkConfig,
 )
-from qoala.runtime.environment import GlobalEnvironment
+from qoala.runtime.environment import NetworkInfo
+from qoala.runtime.lhi import (
+    INSTR_MEASURE_INSTANT,
+    LhiLatencies,
+    LhiLinkInfo,
+    LhiNetworkInfo,
+    LhiProcNodeInfo,
+    LhiTopology,
+    LhiTopologyBuilder,
+)
+from qoala.runtime.lhi_nv_compat import LhiTopologyBuilderForOldNV
+from qoala.runtime.lhi_to_ehi import (
+    GenericToVanillaInterface,
+    LhiConverter,
+    NativeToFlavourInterface,
+    NvToNvInterface,
+)
+from qoala.sim.entdist.entdist import EntDist
+from qoala.sim.entdist.entdistcomp import EntDistComponent
 from qoala.sim.network import ProcNodeNetwork
 from qoala.sim.procnode import ProcNode
+from qoala.util.math import fidelity_to_prob_max_mixed
 
 
-# TODO: move this to somewhere else
-def fidelity_to_prob_max_mixed(fid: float) -> float:
-    return (1 - fid) * 4.0 / 3.0
+def build_qprocessor_from_topology(
+    name: str, topology: LhiTopology
+) -> QuantumProcessor:
+    num_qubits = len(topology.qubit_infos)
+
+    mem_noise_models: List[QuantumErrorModel] = []
+    for i in range(num_qubits):
+        info = topology.qubit_infos[i]
+        noise_model = info.error_model(**info.error_model_kwargs)
+        mem_noise_models.append(noise_model)
+
+    phys_instructions: List[PhysicalInstruction] = []
+    # single-qubit gates
+    for qubit_id, gate_infos in topology.single_gate_infos.items():
+        for gate_info in gate_infos:
+            # TODO: refactor this hack
+            if gate_info.instruction == INSTR_MEASURE_INSTANT:
+                duration = 0.0
+            else:
+                duration = gate_info.duration
+
+            phys_instr = PhysicalInstruction(
+                instruction=gate_info.instruction,
+                duration=duration,
+                topology=[qubit_id],
+                quantum_noise_model=gate_info.error_model(
+                    **gate_info.error_model_kwargs
+                ),
+            )
+            phys_instructions.append(phys_instr)
+
+    # multi-qubit gates
+    for multi_qubit, gate_infos in topology.multi_gate_infos.items():
+        qubit_ids = tuple(multi_qubit.qubit_ids)
+        for gate_info in gate_infos:
+            phys_instr = PhysicalInstruction(
+                instruction=gate_info.instruction,
+                duration=gate_info.duration,
+                topology=[qubit_ids],
+                quantum_noise_model=gate_info.error_model(
+                    **gate_info.error_model_kwargs
+                ),
+            )
+            phys_instructions.append(phys_instr)
+
+    return QuantumProcessor(
+        name=name,
+        num_positions=num_qubits,
+        mem_noise_models=mem_noise_models,
+        phys_instructions=phys_instructions,
+    )
 
 
 def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumProcessor:
@@ -67,7 +138,6 @@ def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumPro
     phys_instructions.append(
         PhysicalInstruction(
             INSTR_INIT,
-            parallel=False,
             duration=cfg.init_time,
         )
     )
@@ -84,9 +154,7 @@ def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumPro
         phys_instructions.append(
             PhysicalInstruction(
                 instr,
-                parallel=False,
                 quantum_noise_model=single_qubit_gate_noise,
-                apply_q_noise_after=True,
                 duration=cfg.single_qubit_gate_time,
             )
         )
@@ -95,16 +163,13 @@ def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumPro
         phys_instructions.append(
             PhysicalInstruction(
                 instr,
-                parallel=False,
                 quantum_noise_model=two_qubit_gate_noise,
-                apply_q_noise_after=True,
                 duration=cfg.two_qubit_gate_time,
             )
         )
 
     phys_instr_measure = PhysicalInstruction(
         INSTR_MEASURE,
-        parallel=False,
         duration=cfg.measure_time,
     )
     phys_instructions.append(phys_instr_measure)
@@ -121,7 +186,6 @@ def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumPro
 
 
 def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
-
     # noise models for single- and multi-qubit operations
     electron_init_noise = DepolarNoiseModel(
         depolar_rate=cfg.electron_init_depolar_prob, time_independent=True
@@ -157,25 +221,21 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
     phys_instructions.append(
         PhysicalInstruction(
             INSTR_INIT,
-            parallel=False,
             topology=carbon_positions,
             quantum_noise_model=carbon_init_noise,
-            apply_q_noise_after=True,
             duration=cfg.carbon_init,
         )
     )
 
-    for (instr, dur) in zip(
+    for instr, dur in zip(
         [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
         [cfg.carbon_rot_x, cfg.carbon_rot_y, cfg.carbon_rot_z],
     ):
         phys_instructions.append(
             PhysicalInstruction(
                 instr,
-                parallel=False,
                 topology=carbon_positions,
                 quantum_noise_model=carbon_z_rot_noise,
-                apply_q_noise_after=True,
                 duration=dur,
             )
         )
@@ -183,25 +243,21 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
     phys_instructions.append(
         PhysicalInstruction(
             INSTR_INIT,
-            parallel=False,
             topology=[electron_position],
             quantum_noise_model=electron_init_noise,
-            apply_q_noise_after=True,
             duration=cfg.electron_init,
         )
     )
 
-    for (instr, dur) in zip(
+    for instr, dur in zip(
         [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
         [cfg.electron_rot_x, cfg.electron_rot_y, cfg.electron_rot_z],
     ):
         phys_instructions.append(
             PhysicalInstruction(
                 instr,
-                parallel=False,
                 topology=[electron_position],
                 quantum_noise_model=electron_single_qubit_noise,
-                apply_q_noise_after=True,
                 duration=dur,
             )
         )
@@ -212,10 +268,8 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
     phys_instructions.append(
         PhysicalInstruction(
             INSTR_CXDIR,
-            parallel=False,
             topology=electron_carbon_topologies,
             quantum_noise_model=ec_noise,
-            apply_q_noise_after=True,
             duration=cfg.ec_controlled_dir_x,
         )
     )
@@ -223,10 +277,8 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
     phys_instructions.append(
         PhysicalInstruction(
             INSTR_CYDIR,
-            parallel=False,
             topology=electron_carbon_topologies,
             quantum_noise_model=ec_noise,
-            apply_q_noise_after=True,
             duration=cfg.ec_controlled_dir_y,
         )
     )
@@ -243,7 +295,6 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
 
     phys_instr_measure = PhysicalInstruction(
         INSTR_MEASURE,
-        parallel=False,
         topology=[electron_position],
         quantum_noise_model=None,
         duration=cfg.measure,
@@ -264,35 +315,38 @@ def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
     return qmem
 
 
-def build_procnode(cfg: ProcNodeConfig, global_env: GlobalEnvironment) -> ProcNode:
-    if cfg.qdevice_typ == "nv":
-        qdevice_cfg = cfg.qdevice_cfg
-        if not isinstance(qdevice_cfg, NVQDeviceConfig):
-            qdevice_cfg = NVQDeviceConfig(**cfg.qdevice_cfg)
-        qprocessor = build_nv_qprocessor(f"qdevice_{cfg.name}", cfg=qdevice_cfg)
-        procnode = ProcNode(
-            cfg.name,
-            global_env=global_env,
-            qprocessor=qprocessor,
-            qdevice_type="nv",
-            node_id=cfg.node_id,
-        )
-    elif cfg.qdevice_typ == "generic":
-        qdevice_cfg = cfg.qdevice_cfg
-        if not isinstance(qdevice_cfg, GenericQDeviceConfig):
-            qdevice_cfg = GenericQDeviceConfig(**cfg.qdevice_cfg)
-        qprocessor = build_generic_qprocessor(f"qdevice_{cfg.name}", cfg=qdevice_cfg)
-        procnode = ProcNode(
-            cfg.name,
-            global_env=global_env,
-            qprocessor=qprocessor,
-            qdevice_type="generic",
-            node_id=cfg.node_id,
-        )
-        # TODO: do this in constructor?
-        procnode.qnos.processor.instr_latency = cfg.instr_latency
-        procnode.host.processor.instr_latency = cfg.instr_latency
-        procnode.host.processor.receive_latency = cfg.receive_latency
+def build_procnode(
+    cfg: ProcNodeConfig, network_info: NetworkInfo, network_ehi: EhiNetworkInfo
+) -> ProcNode:
+    # TODO: Refactor ad-hoc way of old NV config
+    # TODO: Refactor how ntf interface is configured!
+    ntf_interface: NativeToFlavourInterface
+    if cfg.topology is not None:
+        topology = LhiTopologyBuilder.from_config(cfg.topology)
+        ntf_interface = GenericToVanillaInterface()
+    if cfg.nv_config is not None:
+        topology = LhiTopologyBuilderForOldNV.from_nv_config(cfg.nv_config)
+        ntf_interface = NvToNvInterface()
+
+    qprocessor = build_qprocessor_from_topology(name=cfg.node_name, topology=topology)
+    latencies = LhiLatencies.from_config(cfg.latencies)
+    procnode = ProcNode(
+        cfg.node_name,
+        network_info=network_info,
+        qprocessor=qprocessor,
+        qdevice_topology=topology,
+        latencies=latencies,
+        ntf_interface=ntf_interface,
+        node_id=cfg.node_id,
+        network_ehi=network_ehi,
+    )
+
+    # TODO: refactor this hack
+    procnode.qnos.processor._latencies.qnos_instr_time = cfg.latencies.qnos_instr_time
+    procnode.host.processor._latencies.host_instr_time = cfg.latencies.host_instr_time
+    procnode.host.processor._latencies.host_peer_latency = (
+        cfg.latencies.host_peer_latency
+    )
     return procnode
 
 
@@ -305,8 +359,8 @@ def build_ll_protocol(
         )
     elif config.typ == "depolarise":
         link_cfg = config.cfg
-        if not isinstance(link_cfg, DepolariseLinkConfig):
-            link_cfg = DepolariseLinkConfig(**config.cfg)
+        if not isinstance(link_cfg, DepolariseOldLinkConfig):
+            link_cfg = DepolariseOldLinkConfig(**config.cfg)
         prob_max_mixed = fidelity_to_prob_max_mixed(link_cfg.fidelity)
         link_dist = DepolariseWithFailureMagicDistributor(
             nodes=[proc_node1.node, proc_node2.node],
@@ -314,22 +368,23 @@ def build_ll_protocol(
             prob_success=link_cfg.prob_success,
             t_cycle=link_cfg.t_cycle,
         )
-    elif config.typ == "nv":
-        link_cfg = config.cfg
-        if not isinstance(link_cfg, NVLinkConfig):
-            link_cfg = NVLinkConfig(**config.cfg)
-        link_dist = NVSingleClickMagicDistributor(
-            nodes=[proc_node1.node, proc_node2.node],
-            length_A=link_cfg.length_A,
-            length_B=link_cfg.length_B,
-            full_cycle=link_cfg.full_cycle,
-            cycle_time=link_cfg.cycle_time,
-            alpha=link_cfg.alpha,
-        )
+    # TODO: decide whether this is still wanted/needed
+    # elif config.typ == "nv":
+    #     link_cfg = config.cfg
+    #     if not isinstance(link_cfg, NVLinkConfig):
+    #         link_cfg = NVLinkConfig(**config.cfg)
+    #     link_dist = NVSingleClickMagicDistributor(
+    #         nodes=[proc_node1.node, proc_node2.node],
+    #         length_A=link_cfg.length_A,
+    #         length_B=link_cfg.length_B,
+    #         full_cycle=link_cfg.full_cycle,
+    #         cycle_time=link_cfg.cycle_time,
+    #         alpha=link_cfg.alpha,
+    #     )
     elif config.typ == "heralded":
         link_cfg = config.cfg
-        if not isinstance(link_cfg, HeraldedLinkConfig):
-            link_cfg = HeraldedLinkConfig(**config.cfg)
+        if not isinstance(link_cfg, HeraldedOldLinkConfig):
+            link_cfg = HeraldedOldLinkConfig(**config.cfg)
         connection = MiddleHeraldedConnection(name="heralded_conn", **link_cfg.dict())
         link_dist = DoubleClickMagicDistributor(
             [proc_node1.node, proc_node2.node], connection
@@ -345,24 +400,89 @@ def build_ll_protocol(
 
 
 def build_network(
-    config: ProcNodeNetworkConfig, global_env: GlobalEnvironment
+    config: ProcNodeNetworkConfig,
+    network_info: NetworkInfo,
 ) -> ProcNodeNetwork:
-    proc_nodes: Dict[str, ProcNode] = {}
-    link_prots: List[MagicLinkLayerProtocol] = []
+    procnodes: Dict[str, ProcNode] = {}
+
+    ehi_links: Dict[Tuple[int, int], EhiLinkInfo] = {}
+    for link_between_nodes in config.links:
+        lhi_link = LhiLinkInfo.from_config(link_between_nodes.link_config)
+        ehi_link = LhiConverter.link_info_to_ehi(lhi_link)
+        ids = (link_between_nodes.node_id1, link_between_nodes.node_id2)
+        ehi_links[ids] = ehi_link
+    network_ehi = EhiNetworkInfo(ehi_links)
 
     for cfg in config.nodes:
-        proc_nodes[cfg.name] = build_procnode(cfg, global_env)
+        procnodes[cfg.node_name] = build_procnode(cfg, network_info, network_ehi)
 
-    for (_, s1), (_, s2) in itertools.combinations(proc_nodes.items(), 2):
+    ns_nodes = [procnode.node for procnode in procnodes.values()]
+    entdistcomp = EntDistComponent(network_info)
+    entdist = EntDist(nodes=ns_nodes, network_info=network_info, comp=entdistcomp)
+
+    for link_between_nodes in config.links:
+        link = LhiLinkInfo.from_config(link_between_nodes.link_config)
+        n1 = link_between_nodes.node_id1
+        n2 = link_between_nodes.node_id2
+        entdist.add_sampler(n1, n2, link)
+
+    for (_, s1), (_, s2) in itertools.combinations(procnodes.items(), 2):
         s1.connect_to(s2)
 
-    for cfg in config.links:
-        proc_node1 = proc_nodes[cfg.node1]
-        proc_node2 = proc_nodes[cfg.node2]
-        link_prot = build_ll_protocol(cfg, proc_node1, proc_node2)
-        proc_node1.assign_ll_protocol(proc_node2.node.ID, link_prot)
-        proc_node2.assign_ll_protocol(proc_node1.node.ID, link_prot)
+    for name, procnode in procnodes.items():
+        procnode.node.entdist_out_port.connect(entdistcomp.node_in_port(name))
+        procnode.node.entdist_in_port.connect(entdistcomp.node_out_port(name))
 
-        link_prots.append(link_prot)
+    return ProcNodeNetwork(procnodes, entdist)
 
-    return ProcNodeNetwork(proc_nodes, link_prots)
+
+def build_procnode_from_lhi(
+    id: int,
+    name: str,
+    topology: LhiTopology,
+    latencies: LhiLatencies,
+    network_info: NetworkInfo,
+    network_lhi: LhiNetworkInfo,
+) -> ProcNode:
+    qprocessor = build_qprocessor_from_topology(f"{name}_processor", topology)
+    network_ehi = LhiConverter.network_to_ehi(network_lhi)
+    return ProcNode(
+        name=name,
+        node_id=id,
+        network_info=network_info,
+        qprocessor=qprocessor,
+        qdevice_topology=topology,
+        latencies=latencies,
+        ntf_interface=GenericToVanillaInterface(),
+        network_ehi=network_ehi,
+    )
+
+
+def build_network_from_lhi(
+    procnode_infos: List[LhiProcNodeInfo],
+    network_info: NetworkInfo,
+    network_lhi: LhiNetworkInfo,
+) -> ProcNodeNetwork:
+    procnodes: Dict[str, ProcNode] = {}
+
+    for info in procnode_infos:
+        procnode = build_procnode_from_lhi(
+            info.id, info.name, info.topology, info.latencies, network_info, network_lhi
+        )
+        procnodes[info.name] = procnode
+
+    ns_nodes = [procnode.node for procnode in procnodes.values()]
+    entdistcomp = EntDistComponent(network_info)
+    entdist = EntDist(nodes=ns_nodes, network_info=network_info, comp=entdistcomp)
+
+    for ([n1, n2], link_info) in network_lhi.links.items():
+        entdist.add_sampler(n1, n2, link_info)
+
+    for (_, s1), (_, s2) in itertools.combinations(procnodes.items(), 2):
+        s1.connect_to(s2)
+
+    for name, procnode in procnodes.items():
+        procnode.node.entdist_out_port.connect(entdistcomp.node_in_port(name))
+        procnode.node.entdist_in_port.connect(entdistcomp.node_out_port(name))
+
+    return ProcNodeNetwork(procnodes, entdist)
