@@ -1,11 +1,11 @@
 from __future__ import annotations
 
-from typing import Dict, List
+from typing import Dict
 
 from netsquid.components import QuantumProcessor
 from netsquid.nodes.network import Network
-from netsquid_magic.link_layer import MagicLinkLayerProtocol
 
+from qoala.sim.entdist.entdist import EntDist
 from qoala.sim.procnode import ProcNode
 
 
@@ -13,9 +13,7 @@ class ProcNodeNetwork(Network):
     """A network of `ProcNode`s connected by links, which are
     `MagicLinkLayerProtocol`s."""
 
-    def __init__(
-        self, nodes: Dict[str, ProcNode], links: List[MagicLinkLayerProtocol]
-    ) -> None:
+    def __init__(self, nodes: Dict[str, ProcNode], entdist: EntDist) -> None:
         """ProcNodeNetwork constructor.
 
         :param nodes: dictionary of node name to `ProcNode` object representing
@@ -24,15 +22,15 @@ class ProcNodeNetwork(Network):
         contains the IDs of the two nodes that this link connects
         """
         self._nodes = nodes
-        self._links = links
+        self._entdist = entdist
 
     @property
     def nodes(self) -> Dict[str, ProcNode]:
         return self._nodes
 
     @property
-    def links(self) -> List[MagicLinkLayerProtocol]:
-        return self._links
+    def entdist(self) -> EntDist:
+        return self._entdist
 
     @property
     def qdevices(self) -> Dict[str, QuantumProcessor]:
@@ -41,3 +39,10 @@ class ProcNodeNetwork(Network):
     def start_all_nodes(self) -> None:
         for node in self.nodes.values():
             node.start()
+
+    def start_entdist(self) -> None:
+        self.entdist.start()
+
+    def start(self) -> None:
+        self.start_entdist()
+        self.start_all_nodes()
