@@ -1,30 +1,24 @@
 from __future__ import annotations
 
-from typing import Dict, Optional, Type
+from typing import Dict, Optional
 
 from netsquid.components import QuantumProcessor
 from netsquid.protocols import Protocol
-from netsquid_magic.link_layer import MagicLinkLayerProtocolWithSignaling
 
-from qoala.runtime.environment import GlobalEnvironment, LocalEnvironment
+from qoala.lang.ehi import EhiNetworkInfo, EhiNodeInfo
+from qoala.runtime.environment import LocalEnvironment, NetworkInfo
+from qoala.runtime.lhi import LhiLatencies, LhiTopology
+from qoala.runtime.lhi_to_ehi import LhiConverter, NativeToFlavourInterface
 from qoala.runtime.program import BatchInfo, ProgramBatch
-from qoala.runtime.schedule import Schedule, ScheduleSolver
-from qoala.sim.egp import EgpProtocol
-from qoala.sim.egpmgr import EgpManager
-from qoala.sim.host import Host
-from qoala.sim.hostcomp import HostComponent
+from qoala.sim.host.host import Host
+from qoala.sim.host.hostcomp import HostComponent
+from qoala.sim.host.hostinterface import HostLatencies
 from qoala.sim.memmgr import MemoryManager
-from qoala.sim.netstack import Netstack, NetstackComponent
-from qoala.sim.process import IqoalaProcess
+from qoala.sim.netstack import Netstack, NetstackComponent, NetstackLatencies
+from qoala.sim.process import QoalaProcess
 from qoala.sim.procnodecomp import ProcNodeComponent
-from qoala.sim.qdevice import (
-    GenericPhysicalQuantumMemory,
-    NVPhysicalQuantumMemory,
-    QDevice,
-    QDeviceType,
-)
-from qoala.sim.qnos import Qnos
-from qoala.sim.qnoscomp import QnosComponent
+from qoala.sim.qdevice import QDevice
+from qoala.sim.qnos import Qnos, QnosComponent, QnosLatencies
 from qoala.sim.scheduler import Scheduler
 
 
@@ -34,10 +28,13 @@ class ProcNode(Protocol):
     def __init__(
         self,
         name: str,
-        global_env: GlobalEnvironment,
+        network_info: NetworkInfo,
         qprocessor: QuantumProcessor,
+        qdevice_topology: LhiTopology,
+        latencies: LhiLatencies,
+        ntf_interface: NativeToFlavourInterface,
+        network_ehi: EhiNetworkInfo,
         node: Optional[ProcNodeComponent] = None,
-        qdevice_type: Optional[str] = "generic",
         node_id: Optional[int] = None,
         scheduler: Optional[Scheduler] = None,
         asynchronous: bool = False,
@@ -61,39 +58,49 @@ class ProcNode(Protocol):
         if node:
             self._node = node
         else:
-            self._node = ProcNodeComponent(name, qprocessor, global_env, node_id)
+            self._node = ProcNodeComponent(name, qprocessor, network_info, node_id)
 
-        self._global_env = global_env
-        self._local_env = LocalEnvironment(global_env, global_env.get_node_id(name))
+        self._network_info = network_info
+        self._network_ehi = network_ehi
+        self._local_env = LocalEnvironment(network_info, network_info.get_node_id(name))
+        self._ntf_interface = ntf_interface
         self._asynchronous = asynchronous
 
         # Create internal components.
-        self._qdevice: QDevice
-        if qdevice_type == "generic":
-            physical_memory = GenericPhysicalQuantumMemory(qprocessor.num_positions)
-            self._qdevice = QDevice(self._node, QDeviceType.GENERIC, physical_memory)
-        elif qdevice_type == "nv":
-            physical_memory = NVPhysicalQuantumMemory(qprocessor.num_positions)
-            self._qdevice = QDevice(self._node, QDeviceType.NV, physical_memory)
-        else:
-            raise ValueError
+        self._qdevice: QDevice = QDevice(self._node, qdevice_topology)
+        self._local_ehi: EhiNodeInfo = LhiConverter.to_ehi(
+            qdevice_topology, ntf_interface, latencies
+        )
 
-        self._host = Host(self.host_comp, self._local_env, self._asynchronous)
-        self._memmgr = MemoryManager(self.node.name, self._qdevice)
-        self._egpmgr = EgpManager()
+        host_latencies = HostLatencies(
+            latencies.host_instr_time,
+            latencies.host_peer_latency,
+        )
+        qnos_latencies = QnosLatencies(
+            latencies.qnos_instr_time,
+        )
+        # TODO: decide if still needed
+        netstack_latencies = NetstackLatencies(0)
+
+        self._host = Host(
+            self.host_comp, self._local_env, host_latencies, self._asynchronous
+        )
+        self._memmgr = MemoryManager(self.node.name, self._qdevice, self._local_ehi)
         self._qnos = Qnos(
             self.qnos_comp,
             self._local_env,
             self._memmgr,
             self._qdevice,
+            qnos_latencies,
+            self._ntf_interface,
             self._asynchronous,
         )
         self._netstack = Netstack(
             self.netstack_comp,
             self._local_env,
             self._memmgr,
-            self._egpmgr,
             self._qdevice,
+            netstack_latencies,
         )
 
         if scheduler is None:
@@ -104,22 +111,12 @@ class ProcNode(Protocol):
                 self._netstack,
                 self._memmgr,
                 self._local_env,
+                self._local_ehi,
+                self._network_ehi,
             )
         else:
             self._scheduler = scheduler
 
-    def install_schedule(self, schedule: Schedule) -> None:
-        self.scheduler.install_schedule(schedule)
-
-    def assign_ll_protocol(
-        self, remote_id: int, prot: MagicLinkLayerProtocolWithSignaling
-    ) -> None:
-        """Set the link layer protocol to use for entanglement generation.
-
-        The same link layer protocol object is used by both nodes sharing a link in
-        the network."""
-        self.egpmgr.add_egp(remote_id, EgpProtocol(self.node, prot))
-
     @property
     def node(self) -> ProcNodeComponent:
         return self._node
@@ -141,7 +138,7 @@ class ProcNode(Protocol):
         return self._qdevice
 
     @qdevice.setter
-    def qdevice(self, qdevice) -> QDevice:
+    def qdevice(self, qdevice) -> None:
         self._qdevice = qdevice
         self.qnos.qdevice = qdevice
         self.netstack.qdevice = qdevice
@@ -174,10 +171,6 @@ class ProcNode(Protocol):
     def memmgr(self) -> MemoryManager:
         return self._memmgr
 
-    @property
-    def egpmgr(self) -> EgpManager:
-        return self._egpmgr
-
     @property
     def scheduler(self) -> Scheduler:
         return self._scheduler
@@ -186,6 +179,22 @@ class ProcNode(Protocol):
     def scheduler(self, scheduler: Scheduler) -> None:
         self._scheduler = scheduler
 
+    @property
+    def local_ehi(self) -> EhiNodeInfo:
+        return self._local_ehi
+
+    @local_ehi.setter
+    def local_ehi(self, local_ehi: EhiNodeInfo) -> None:
+        self._local_ehi = local_ehi
+
+    @property
+    def network_ehi(self) -> EhiNetworkInfo:
+        return self._network_ehi
+
+    @network_ehi.setter
+    def network_ehi(self, network_ehi: EhiNetworkInfo) -> None:
+        self._network_ehi = network_ehi
+
     def connect_to(self, other: ProcNode) -> None:
         """Create connections between ports of this ProcNode and those of
         another ProcNode."""
@@ -226,10 +235,7 @@ class ProcNode(Protocol):
     def initialize_processes(self) -> None:
         self.scheduler.create_processes_for_batches()
 
-    def initialize_schedule(self, solver: Type[ScheduleSolver]) -> None:
-        self.scheduler.solve_and_install_schedule(solver)
-
-    def add_process(self, process: IqoalaProcess) -> None:
+    def add_process(self, process: QoalaProcess) -> None:
         self.memmgr.add_process(process)
 
     def get_batches(self) -> Dict[int, ProgramBatch]:
