@@ -5,27 +5,31 @@ from netqasm.lang.instr.core import MeasInstruction, SetInstruction
 from netqasm.lang.operand import Register, Template
 from netqasm.lang.subroutine import Subroutine
 
-from qoala.lang.iqoala import (
+from qoala.lang.hostlang import (
     AssignCValueOp,
-    IqoalaInstrParser,
-    IqoalaMetaParser,
-    IqoalaParseError,
-    IqoalaParser,
-    IqoalaRequest,
-    IQoalaRequestParser,
-    IqoalaSharedMemLoc,
-    IqoalaSubroutine,
-    IQoalaSubroutineParser,
+    BasicBlockType,
     IqoalaVector,
-    ProgramMeta,
     RunSubroutineOp,
 )
-from qoala.sim.requests import (
-    EprCreateRole,
-    EprCreateType,
-    NetstackCreateRequest,
-    NetstackReceiveRequest,
+from qoala.lang.parse import (
+    HostCodeParser,
+    IqoalaInstrParser,
+    IqoalaMetaParser,
+    LocalRoutineParser,
+    QoalaParseError,
+    QoalaParser,
+    RequestRoutineParser,
 )
+from qoala.lang.program import LocalRoutine, ProgramMeta
+from qoala.lang.request import (
+    CallbackType,
+    EprRole,
+    EprType,
+    QoalaRequest,
+    RequestRoutine,
+    RequestVirtIdMapping,
+)
+from qoala.lang.routine import RoutineMetadata
 from qoala.util.tests import text_equal
 
 
@@ -38,7 +42,7 @@ csockets: 0 -> bob
 META_END
     """
 
-    with pytest.raises(IqoalaParseError):
+    with pytest.raises(QoalaParseError):
         IqoalaMetaParser(text).parse()
 
 
@@ -51,7 +55,7 @@ csockets: 0 -> bob
 epr_sockets: 
     """
 
-    with pytest.raises(IqoalaParseError):
+    with pytest.raises(QoalaParseError):
         IqoalaMetaParser(text).parse()
 
 
@@ -121,7 +125,7 @@ def test_parse_faulty_instr():
 x = assign_cval
     """
 
-    with pytest.raises(IqoalaParseError):
+    with pytest.raises(QoalaParseError):
         IqoalaInstrParser(text).parse()
 
 
@@ -175,18 +179,84 @@ vec<m1; m2> = run_subroutine(vec<x; y>) : subrt1
     )
 
 
+def test_parse_block_header():
+    text = "^b0 {type = CL}:"
+
+    name, typ = HostCodeParser("")._parse_block_header(text)
+    assert name == "b0"
+    assert typ == BasicBlockType.CL
+
+
+def test_parse_block():
+    text = """
+^b0 {type = CL}:
+    x = assign_cval() : 1
+    y = assign_cval() : 17
+    """
+
+    block = HostCodeParser("").parse_block(text)
+    assert block.name == "b0"
+    assert block.typ == BasicBlockType.CL
+    assert len(block.instructions) == 2
+    assert block.instructions[0] == AssignCValueOp(result="x", value=1)
+    assert block.instructions[1] == AssignCValueOp(result="y", value=17)
+
+
+def test_get_block_texts():
+    text = """
+^b0 {type = CL}:
+    x = assign_cval() : 1
+    y = assign_cval() : 17
+
+^b1 {type = QL}:
+    run_subroutine(vec<x>) : subrt1
+    """
+
+    block_texts = HostCodeParser(text).get_block_texts()
+    assert len(block_texts) == 2
+
+
+def test_parse_multiple_blocks():
+    text = """
+^b0 {type = CL}:
+    x = assign_cval() : 1
+    y = assign_cval() : 17
+
+^b1 {type = QL}:
+    run_subroutine(vec<x>) : subrt1
+    """
+
+    blocks = HostCodeParser(text).parse()
+    assert len(blocks) == 2
+
+    assert blocks[0].name == "b0"
+    assert blocks[0].typ == BasicBlockType.CL
+    assert len(blocks[0].instructions) == 2
+    assert blocks[0].instructions[0] == AssignCValueOp(result="x", value=1)
+    assert blocks[0].instructions[1] == AssignCValueOp(result="y", value=17)
+
+    assert blocks[1].name == "b1"
+    assert blocks[1].typ == BasicBlockType.QL
+    assert len(blocks[1].instructions) == 1
+    assert blocks[1].instructions[0] == RunSubroutineOp(
+        result=None, values=IqoalaVector(["x"]), subrt="subrt1"
+    )
+
+
 def test_parse_subrt():
     text = """
 SUBROUTINE subrt1
     params: my_value
-    returns: M0 -> m
+    returns: m
+    uses: 
+    keeps:
     request: 
   NETQASM_START
     set Q0 {my_value}
   NETQASM_END
     """
 
-    parsed = IQoalaSubroutineParser(text).parse()
+    parsed = LocalRoutineParser(text).parse()
     assert len(parsed) == 1
     assert "subrt1" in parsed
     subrt = parsed["subrt1"]
@@ -195,10 +265,11 @@ SUBROUTINE subrt1
         SetInstruction(reg=Register.from_str("Q0"), imm=Template("my_value"))
     ]
     expected_args = ["my_value"]
-    assert subrt == IqoalaSubroutine(
+    assert subrt == LocalRoutine(
         name="subrt1",
         subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
-        return_map={"m": IqoalaSharedMemLoc("M0")},
+        metadata=RoutineMetadata.use_none(),
+        return_vars=["m"],
     )
 
 
@@ -206,7 +277,9 @@ def test_parse_subrt_2():
     text = """
 SUBROUTINE my_subroutine
     params: param1, param2
-    returns: M5 -> result1, M6 -> result2
+    returns: result1, result2
+    uses: 0, 1
+    keeps: 
     request: 
   NETQASM_START
     set R0 {param1}
@@ -216,7 +289,7 @@ SUBROUTINE my_subroutine
   NETQASM_END
     """
 
-    parsed = IQoalaSubroutineParser(text).parse()
+    parsed = LocalRoutineParser(text).parse()
     assert len(parsed) == 1
     assert "my_subroutine" in parsed
     subrt = parsed["my_subroutine"]
@@ -228,13 +301,11 @@ SUBROUTINE my_subroutine
         MeasInstruction(reg0=Register.from_str("Q1"), reg1=Register.from_str("M6")),
     ]
     expected_args = ["param1", "param2"]
-    assert subrt == IqoalaSubroutine(
+    assert subrt == LocalRoutine(
         name="my_subroutine",
         subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
-        return_map={
-            "result1": IqoalaSharedMemLoc("M5"),
-            "result2": IqoalaSharedMemLoc("M6"),
-        },
+        metadata=RoutineMetadata.free_all([0, 1]),
+        return_vars=["result1", "result2"],
     )
 
 
@@ -242,7 +313,9 @@ def test_parse_multiple_subrt():
     text = """
 SUBROUTINE subrt1
     params: param1
-    returns: M0 -> m
+    returns: m
+    uses: 0
+    keeps:
     request: 
   NETQASM_START
     set R0 {param1}
@@ -252,13 +325,15 @@ SUBROUTINE subrt1
 SUBROUTINE subrt2
     params: theta
     returns: 
+    uses: 
+    keeps:
     request: 
   NETQASM_START
     set R0 {theta}
   NETQASM_END
     """
 
-    parsed = IQoalaSubroutineParser(text).parse()
+    parsed = LocalRoutineParser(text).parse()
     assert len(parsed) == 2
     assert "subrt1" in parsed
     assert "subrt2" in parsed
@@ -275,17 +350,17 @@ SUBROUTINE subrt2
     ]
     expected_args_2 = ["theta"]
 
-    assert subrt1 == IqoalaSubroutine(
+    assert subrt1 == LocalRoutine(
         name="subrt1",
         subrt=Subroutine(instructions=expected_instrs_1, arguments=expected_args_1),
-        return_map={
-            "m": IqoalaSharedMemLoc("M0"),
-        },
+        metadata=RoutineMetadata.free_all([0]),
+        return_vars=["m"],
     )
-    assert subrt2 == IqoalaSubroutine(
+    assert subrt2 == LocalRoutine(
         name="subrt2",
         subrt=Subroutine(instructions=expected_instrs_2, arguments=expected_args_2),
-        return_map={},
+        metadata=RoutineMetadata.use_none(),
+        return_vars=[],
     )
 
 
@@ -293,46 +368,53 @@ def test_parse_invalid_subrt():
     text = """
 SUBROUTINE my_subroutine
     params: param1, param2
-    returns: M5 -> result1, M6 -> result2
+    returns: result1, result2
     request: 
   NETQASM_START
     set R0 {param3}
   NETQASM_END
     """
 
-    with pytest.raises(IqoalaParseError):
-        IQoalaSubroutineParser(text).parse()
+    with pytest.raises(QoalaParseError):
+        LocalRoutineParser(text).parse()
 
 
 def test_parse_request():
     text = """
 REQUEST req1
-  role: create
+  callback_type: wait_all
+  callback: 
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: create_keep
   num_pairs: 5
+  virt_ids: all 0
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 0, 0, 0, 0, 0
-  result_array_addr: 3
+  typ: create_keep
+  role: create
     """
 
-    parsed = IQoalaRequestParser(text).parse()
+    parsed = RequestRoutineParser(text).parse()
     assert len(parsed) == 1
     assert "req1" in parsed
-    request = parsed["req1"]
+    routine = parsed["req1"]
 
-    assert request == IqoalaRequest(
+    assert routine == RequestRoutine(
         name="req1",
-        role=EprCreateRole.CREATE,
-        request=NetstackCreateRequest(
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
             remote_id=1,
             epr_socket_id=0,
-            typ=EprCreateType.CREATE_KEEP,
             num_pairs=5,
+            virt_ids=RequestVirtIdMapping.from_str("all 0"),
+            timeout=1000,
             fidelity=0.65,
-            virt_qubit_ids=[0, 0, 0, 0, 0],
-            result_array_addr=3,
+            typ=EprType.CREATE_KEEP,
+            role=EprRole.CREATE,
         ),
     )
 
@@ -340,32 +422,39 @@ REQUEST req1
 def test_parse_request_2():
     text = """
 REQUEST req1
-  role: receive
+  callback_type: sequential
+  callback: subrt1
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: measure_directly
   num_pairs: 3
+  virt_ids: custom 1, 2, 3
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 1, 2, 3
-  result_array_addr: 0
+  typ: measure_directly
+  role: receive
     """
 
-    parsed = IQoalaRequestParser(text).parse()
+    parsed = RequestRoutineParser(text).parse()
     assert len(parsed) == 1
     assert "req1" in parsed
-    request = parsed["req1"]
+    routine = parsed["req1"]
 
-    assert request == IqoalaRequest(
+    assert routine == RequestRoutine(
         name="req1",
-        role=EprCreateRole.RECEIVE,
-        request=NetstackReceiveRequest(
+        return_vars=[],
+        callback_type=CallbackType.SEQUENTIAL,
+        callback="subrt1",
+        request=QoalaRequest(
+            name="req1",
             remote_id=1,
             epr_socket_id=0,
-            typ=EprCreateType.MEASURE_DIRECTLY,
             num_pairs=3,
+            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
+            timeout=1000,
             fidelity=0.65,
-            virt_qubit_ids=[1, 2, 3],
-            result_array_addr=0,
+            typ=EprType.MEASURE_DIRECTLY,
+            role=EprRole.RECEIVE,
         ),
     )
 
@@ -373,32 +462,39 @@ REQUEST req1
 def test_parse_request_with_template():
     text = """
 REQUEST req1
-  role: receive
+  callback_type: wait_all
+  callback: 
+  return_vars: 
   remote_id: {client_id}
   epr_socket_id: 0
-  typ: measure_directly
   num_pairs: 3
+  virt_ids: custom 1, 2, 3
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 1, 2, 3
-  result_array_addr: 0
+  typ: measure_directly
+  role: receive
     """
 
-    parsed = IQoalaRequestParser(text).parse()
+    parsed = RequestRoutineParser(text).parse()
     assert len(parsed) == 1
     assert "req1" in parsed
-    request = parsed["req1"]
+    routine = parsed["req1"]
 
-    assert request == IqoalaRequest(
+    assert routine == RequestRoutine(
         name="req1",
-        role=EprCreateRole.RECEIVE,
-        request=NetstackReceiveRequest(
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
             remote_id=Template("client_id"),
             epr_socket_id=0,
-            typ=EprCreateType.MEASURE_DIRECTLY,
             num_pairs=3,
+            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
+            timeout=1000,
             fidelity=0.65,
-            virt_qubit_ids=[1, 2, 3],
-            result_array_addr=0,
+            typ=EprType.MEASURE_DIRECTLY,
+            role=EprRole.RECEIVE,
         ),
     )
 
@@ -406,57 +502,71 @@ REQUEST req1
 def test_parse_multiple_request():
     text = """
 REQUEST req1
-  role: create
+  callback_type: wait_all
+  callback: 
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: create_keep
   num_pairs: 5
+  virt_ids: all 0
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 0, 0, 0, 0, 0
-  result_array_addr: 3
+  typ: create_keep
+  role: create
 
 REQUEST req2
-  role: receive
+  callback_type: sequential
+  callback: subrt1
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: measure_directly
   num_pairs: 3
+  virt_ids: increment 1
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 1, 2, 3
-  result_array_addr: 0
+  typ: measure_directly
+  role: receive
     """
 
-    parsed = IQoalaRequestParser(text).parse()
+    parsed = RequestRoutineParser(text).parse()
     assert len(parsed) == 2
     assert "req1" in parsed
     assert "req2" in parsed
-    req1 = parsed["req1"]
-    req2 = parsed["req2"]
+    routine1 = parsed["req1"]
+    routine2 = parsed["req2"]
 
-    assert req1 == IqoalaRequest(
+    assert routine1 == RequestRoutine(
         name="req1",
-        role=EprCreateRole.CREATE,
-        request=NetstackCreateRequest(
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
             remote_id=1,
             epr_socket_id=0,
-            typ=EprCreateType.CREATE_KEEP,
             num_pairs=5,
+            virt_ids=RequestVirtIdMapping.from_str("all 0"),
+            timeout=1000,
             fidelity=0.65,
-            virt_qubit_ids=[0, 0, 0, 0, 0],
-            result_array_addr=3,
+            typ=EprType.CREATE_KEEP,
+            role=EprRole.CREATE,
         ),
     )
-    assert req2 == IqoalaRequest(
+    assert routine2 == RequestRoutine(
         name="req2",
-        role=EprCreateRole.RECEIVE,
-        request=NetstackReceiveRequest(
+        return_vars=[],
+        callback_type=CallbackType.SEQUENTIAL,
+        callback="subrt1",
+        request=QoalaRequest(
+            name="req2",
             remote_id=1,
             epr_socket_id=0,
-            typ=EprCreateType.MEASURE_DIRECTLY,
             num_pairs=3,
+            virt_ids=RequestVirtIdMapping.from_str("increment 1"),
+            timeout=1000,
             fidelity=0.65,
-            virt_qubit_ids=[1, 2, 3],
-            result_array_addr=0,
+            typ=EprType.MEASURE_DIRECTLY,
+            role=EprRole.RECEIVE,
         ),
     )
 
@@ -464,20 +574,23 @@ REQUEST req2
 def test_parse_invalid_request():
     text = """
 REQUEST req1
-  role: receive
+  callback_type: wait_all
+  callback: 
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: measure
   num_pairs: 3
+  virt_ids: all 0
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 1, 2, 3
-  result_array_addr: 0
+  typ: invalid
+  role: receive
     """
 
     # invalid 'typ' value
 
-    with pytest.raises(IqoalaParseError):
-        IQoalaRequestParser(text).parse()
+    with pytest.raises(QoalaParseError):
+        RequestRoutineParser(text).parse()
 
 
 DEFAULT_META = """
@@ -501,20 +614,30 @@ META_END
     """
 
     program_text = """
-my_value = assign_cval() : 1
-remote_id = assign_cval() : 0
-send_cmsg(remote_id, my_value)
-received_value = recv_cmsg(remote_id)
-new_value = assign_cval() : 3
-my_value = add_cval_c(new_value, new_value)
-vec<m> = run_subroutine(vec<my_value>) : subrt1
-return_result(m)
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+    remote_id = assign_cval() : 0
+    send_cmsg(remote_id, my_value)
+    received_value = recv_cmsg(remote_id)
+    new_value = assign_cval() : 3
+    my_value = add_cval_c(new_value, new_value)
+
+^b1 {type = QL}:
+    vec<m> = run_subroutine(vec<my_value>) : subrt1
+
+^b2 {type = QC}:
+    vec<m> = run_request(vec<my_value>) : req1
+
+^b3 {type = CL}:
+    return_result(m)
     """
 
     subrt_text = """
 SUBROUTINE subrt1
     params: my_value
-    returns: M0 -> m
+    returns: m
+    uses: 0
+    keeps:
     request: 
   NETQASM_START
     set Q0 0
@@ -526,26 +649,30 @@ SUBROUTINE subrt1
 
     req_text = """
 REQUEST req1
-  role: receive
+  callback_type: wait_all
+  callback: 
+  return_vars: 
   remote_id: 1
   epr_socket_id: 0
-  typ: measure_directly
   num_pairs: 3
+  virt_ids: increment 1
+  timeout: 1000
   fidelity: 0.65
-  virt_qubit_ids: 1, 2, 3
-  result_array_addr: 0
+  typ: measure_directly
+  role: receive
     """
 
-    parsed_program = IqoalaParser(
+    parsed_program = QoalaParser(
         meta_text=meta_text,
-        instr_text=program_text,
+        host_text=program_text,
         subrt_text=subrt_text,
         req_text=req_text,
     ).parse()
-    assert len(parsed_program.instructions) == 8
-    assert "subrt1" in parsed_program.subroutines
-    assert len(parsed_program.requests) == 1
-    assert "req1" in parsed_program.requests
+    assert len(parsed_program.blocks) == 4
+    assert len(parsed_program.instructions) == 9
+    assert "subrt1" in parsed_program.local_routines
+    assert len(parsed_program.request_routines) == 1
+    assert "req1" in parsed_program.request_routines
 
 
 def test_parse_program_no_requests():
@@ -559,20 +686,27 @@ META_END
     """
 
     program_text = """
-my_value = assign_cval() : 1
-remote_id = assign_cval() : 0
-send_cmsg(remote_id, my_value)
-received_value = recv_cmsg(remote_id)
-new_value = assign_cval() : 3
-my_value = add_cval_c(new_value, new_value)
-vec<m> = run_subroutine(vec<my_value>) : subrt1
-return_result(m)
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+    remote_id = assign_cval() : 0
+    send_cmsg(remote_id, my_value)
+    received_value = recv_cmsg(remote_id)
+    new_value = assign_cval() : 3
+    my_value = add_cval_c(new_value, new_value)
+
+^b1 {type = QL}:
+    vec<m> = run_subroutine(vec<my_value>) : subrt1
+
+^b1 {type = CL}:
+    return_result(m)
     """
 
     subrt_text = """
 SUBROUTINE subrt1
     params: my_value
-    returns: M0 -> m
+    returns: m
+    uses: 0
+    keeps:
     request: 
   NETQASM_START
     set Q0 0
@@ -585,15 +719,16 @@ SUBROUTINE subrt1
     req_text = """
     """
 
-    parsed_program = IqoalaParser(
+    parsed_program = QoalaParser(
         meta_text=meta_text,
-        instr_text=program_text,
+        host_text=program_text,
         subrt_text=subrt_text,
         req_text=req_text,
     ).parse()
+    assert len(parsed_program.blocks) == 3
     assert len(parsed_program.instructions) == 8
-    assert "subrt1" in parsed_program.subroutines
-    assert len(parsed_program.requests) == 0
+    assert "subrt1" in parsed_program.local_routines
+    assert len(parsed_program.request_routines) == 0
 
 
 def test_parse_program_no_subroutines():
@@ -607,13 +742,14 @@ META_END
     """
 
     program_text = """
-my_value = assign_cval() : 1
-remote_id = assign_cval() : 0
-send_cmsg(remote_id, my_value)
-received_value = recv_cmsg(remote_id)
-new_value = assign_cval() : 3
-m = add_cval_c(new_value, new_value)
-return_result(m)
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+    remote_id = assign_cval() : 0
+    send_cmsg(remote_id, my_value)
+    received_value = recv_cmsg(remote_id)
+    new_value = assign_cval() : 3
+    m = add_cval_c(new_value, new_value)
+    return_result(m)
     """
 
     subrt_text = """
@@ -622,14 +758,15 @@ return_result(m)
     req_text = """
     """
 
-    parsed_program = IqoalaParser(
+    parsed_program = QoalaParser(
         meta_text=meta_text,
-        instr_text=program_text,
+        host_text=program_text,
         subrt_text=subrt_text,
         req_text=req_text,
     ).parse()
+    assert len(parsed_program.blocks) == 1
     assert len(parsed_program.instructions) == 7
-    assert len(parsed_program.requests) == 0
+    assert len(parsed_program.request_routines) == 0
 
 
 def test_parse_program_invalid_subrt_reference():
@@ -643,25 +780,70 @@ META_END
     """
 
     program_text = """
-my_value = assign_cval() : 1
-vec<m> = run_subroutine(vec<my_value>) : non_existing_subrt
-return_result(m)
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+^b1 {type = QL}:
+    vec<m> = run_subroutine(vec<my_value>) : non_existing_subrt
+^b2 {type = CL}:
+    return_result(m)
     """
 
     subrt_text = """
 SUBROUTINE subrt1
     params: my_value
-    returns: M0 -> m
+    returns: m
+    uses: 
+    keeps:
     request: 
   NETQASM_START
     set R0 0
   NETQASM_END
     """
 
-    with pytest.raises(IqoalaParseError):
-        IqoalaParser(
+    with pytest.raises(QoalaParseError):
+        QoalaParser(
             meta_text=meta_text,
-            instr_text=program_text,
+            host_text=program_text,
+            subrt_text=subrt_text,
+            req_text="",
+        ).parse()
+
+
+def test_parse_program_invalid_req_routine_reference():
+    meta_text = """
+META_START
+name: alice
+parameters: 
+csockets: 0 -> bob
+epr_sockets: 
+META_END
+    """
+
+    program_text = """
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+^b1 {type = QL}:
+    vec<m> = run_request(vec<my_value>) : non_existing_req
+^b2 {type = CL}:
+    return_result(m)
+    """
+
+    subrt_text = """
+SUBROUTINE subrt1
+    params: my_value
+    returns: m
+    uses: 
+    keeps:
+    request: 
+  NETQASM_START
+    set R0 0
+  NETQASM_END
+    """
+
+    with pytest.raises(QoalaParseError):
+        QoalaParser(
+            meta_text=meta_text,
+            host_text=program_text,
             subrt_text=subrt_text,
             req_text="",
         ).parse()
@@ -677,9 +859,10 @@ epr_sockets:
 META_END
     """
 
-    instr_text = """
-m = assign_cval() : 1
-return_result(m)
+    host_text = """
+^b0 {type = CL}:
+    m = assign_cval() : 1
+    return_result(m)
     """
 
     subrt_text = """
@@ -692,12 +875,28 @@ SUBROUTINE subrt1
   NETQASM_END
     """
 
-    text = meta_text + instr_text + subrt_text
-    parser = IqoalaParser(text)
+    req_text = """
+REQUEST req1
+  callback_type: wait_all
+  callback:
+  return_vars: 
+  remote_id: {client_id}
+  epr_socket_id: 0
+  num_pairs: 1
+  virt_ids: all 0
+  timeout:1000
+  fidelity: 1.0
+  typ: create_keep
+  role: receive
+    """
+
+    text = meta_text + host_text + subrt_text + req_text
+    parser = QoalaParser(text)
 
     assert text_equal(parser._meta_text, meta_text)
-    assert text_equal(parser._instr_text, instr_text)
+    assert text_equal(parser._host_text, host_text)
     assert text_equal(parser._subrt_text, subrt_text)
+    assert text_equal(parser._req_text, req_text)
 
 
 def test_split_text_multiple_subroutines():
@@ -710,15 +909,18 @@ epr_sockets:
 META_END
     """
 
-    instr_text = """
-m = assign_cval() : 1
-return_result(m)
+    host_text = """
+^b0 {type = CL}:
+    m = assign_cval() : 1
+    return_result(m)
     """
 
     subrt_text = """
 SUBROUTINE subrt1
     params: 
     returns: 
+    uses:
+    keeps:
     request: 
   NETQASM_START
     set Q0 0
@@ -727,20 +929,62 @@ SUBROUTINE subrt1
 SUBROUTINE subrt2
     params: 
     returns: 
+    uses:
+    keeps:
     request: 
   NETQASM_START
     set Q7 7
   NETQASM_END
     """
 
-    text = meta_text + instr_text + subrt_text
-    parser = IqoalaParser(text)
+    text = meta_text + host_text + subrt_text
+    parser = QoalaParser(text)
 
     assert text_equal(parser._meta_text, meta_text)
-    assert text_equal(parser._instr_text, instr_text)
+    assert text_equal(parser._host_text, host_text)
     assert text_equal(parser._subrt_text, subrt_text)
 
 
+def test_split_text_no_subroutines():
+    meta_text = """
+META_START
+    name: server
+    parameters: client_id
+    csockets: 0 -> client
+    epr_sockets: 0 -> client
+META_END
+    """
+
+    host_text = """
+    ^b0 {type = CL}:
+        remote_id = assign_cval() : {client_id}
+    ^b1 {type = QC}:
+        run_request(vec<>) : req1
+        """
+
+    req_text = """
+REQUEST req1
+    callback_type: wait_all
+    callback:
+  return_vars: 
+    remote_id: {client_id}
+    epr_socket_id: 0
+    num_pairs: 1
+    virt_ids: all 0
+    timeout:1000
+    fidelity: 1.0
+    typ: create_keep
+    role: receive
+        """
+
+    text = meta_text + host_text + req_text
+    parser = QoalaParser(text)
+
+    assert text_equal(parser._meta_text, meta_text)
+    assert text_equal(parser._host_text, host_text)
+    assert text_equal(parser._req_text, req_text)
+
+
 def test_parse_program_single_text():
     text = """
 META_START
@@ -750,18 +994,23 @@ csockets: 0 -> bob
 epr_sockets: 
 META_END
 
-my_value = assign_cval() : 1
-remote_id = assign_cval() : 0
-send_cmsg(remote_id, my_value)
-received_value = recv_cmsg(remote_id)
-new_value = assign_cval() : 3
-my_value = add_cval_c(new_value, new_value)
-vec<m> = run_subroutine(vec<my_value>) : subrt1
-return_result(m)
+^b0 {type = CL}:
+    my_value = assign_cval() : 1
+    remote_id = assign_cval() : 0
+    send_cmsg(remote_id, my_value)
+    received_value = recv_cmsg(remote_id)
+    new_value = assign_cval() : 3
+    my_value = add_cval_c(new_value, new_value)
+^b1 {type = QL}:
+    vec<m> = run_subroutine(vec<my_value>) : subrt1
+^b2 {type = CL}:
+    return_result(m)
 
 SUBROUTINE subrt1
     params: my_value
-    returns: M0 -> m
+    returns: m
+    uses: 0
+    keeps:
     request: 
   NETQASM_START
     set Q0 0
@@ -771,42 +1020,45 @@ SUBROUTINE subrt1
   NETQASM_END
     """
 
-    parsed_program = IqoalaParser(text).parse()
+    parsed_program = QoalaParser(text).parse()
+    assert len(parsed_program.blocks) == 3
     assert len(parsed_program.instructions) == 8
-    assert "subrt1" in parsed_program.subroutines
+    assert "subrt1" in parsed_program.local_routines
 
 
 def test_parse_file():
-    path = os.path.join(os.path.dirname(__file__), "test_bqc_server.iqoala")
+    path = os.path.join(os.path.dirname(__file__), "test_parse_server.iqoala")
     with open(path) as file:
         text = file.read()
-    parsed_program = IqoalaParser(text).parse()
+    parsed_program = QoalaParser(text).parse()
+    assert len(parsed_program.blocks) == 9
     assert len(parsed_program.instructions) == 11
-    assert "create_epr_0" in parsed_program.subroutines
-    assert "create_epr_1" in parsed_program.subroutines
-    assert "local_cphase" in parsed_program.subroutines
-    assert "meas_qubit_0" in parsed_program.subroutines
-    assert "meas_qubit_1" in parsed_program.subroutines
-    assert "req0" in parsed_program.requests
-    assert "req1" in parsed_program.requests
-    assert parsed_program.subroutines["create_epr_0"].request_name == "req0"
-    assert parsed_program.subroutines["create_epr_1"].request_name == "req1"
+    assert "create_epr_0" in parsed_program.local_routines
+    assert "create_epr_1" in parsed_program.local_routines
+    assert "local_cphase" in parsed_program.local_routines
+    assert "meas_qubit_0" in parsed_program.local_routines
+    assert "meas_qubit_1" in parsed_program.local_routines
+    assert "req0" in parsed_program.request_routines
+    assert "req1" in parsed_program.request_routines
+    assert parsed_program.local_routines["create_epr_0"].request_name == "req0"
+    assert parsed_program.local_routines["create_epr_1"].request_name == "req1"
 
 
 def test_parse_file_2():
-    path = os.path.join(os.path.dirname(__file__), "test_bqc_client.iqoala")
+    path = os.path.join(os.path.dirname(__file__), "test_parse_client.iqoala")
     with open(path) as file:
         text = file.read()
-    parsed_program = IqoalaParser(text).parse()
+    parsed_program = QoalaParser(text).parse()
+    assert len(parsed_program.blocks) == 6
     assert len(parsed_program.instructions) == 19
-    assert "create_epr_0" in parsed_program.subroutines
-    assert "post_epr_0" in parsed_program.subroutines
-    assert "create_epr_1" in parsed_program.subroutines
-    assert "post_epr_1" in parsed_program.subroutines
-    assert "req0" in parsed_program.requests
-    assert "req1" in parsed_program.requests
-    assert parsed_program.subroutines["create_epr_0"].request_name == "req0"
-    assert parsed_program.subroutines["create_epr_1"].request_name == "req1"
+    assert "create_epr_0" in parsed_program.local_routines
+    assert "post_epr_0" in parsed_program.local_routines
+    assert "create_epr_1" in parsed_program.local_routines
+    assert "post_epr_1" in parsed_program.local_routines
+    assert "req0" in parsed_program.request_routines
+    assert "req1" in parsed_program.request_routines
+    assert parsed_program.local_routines["create_epr_0"].request_name == "req0"
+    assert parsed_program.local_routines["create_epr_1"].request_name == "req1"
 
 
 if __name__ == "__main__":
@@ -821,6 +1073,10 @@ if __name__ == "__main__":
     test_parse_vec_2_elements()
     test_parse_vec_2_elements_and_return()
     test_parse_vec_2_elements_and_return_2_elements()
+    test_parse_block_header()
+    test_parse_block()
+    test_get_block_texts()
+    test_parse_multiple_blocks()
     test_parse_subrt()
     test_parse_subrt_2()
     test_parse_multiple_subrt()
@@ -834,8 +1090,10 @@ if __name__ == "__main__":
     test_parse_program_no_requests()
     test_parse_program_no_subroutines()
     test_parse_program_invalid_subrt_reference()
+    test_parse_program_invalid_req_routine_reference()
     test_split_text()
     test_split_text_multiple_subroutines()
+    test_split_text_no_subroutines()
     test_parse_program_single_text()
     test_parse_file()
     test_parse_file_2()
